@@ -1,6 +1,6 @@
 /*
     scardcontrol.c: sample code to use/test SCardControl() API
-    Copyright (C) 2004-2011   Ludovic Rousseau
+    Copyright (C) 2004-2019   Ludovic Rousseau
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -72,7 +72,7 @@ else \
 
 #define PCSC_ERROR_CONT(rv, text) \
 if (rv != SCARD_S_SUCCESS) \
-	printf(text ": " BLUE "%s (0x%"LF"X)\n" NORMAL, pcsc_stringify_error(rv), rv); \
+	printf(text ": " RED "%s (0x%"LF"X)\n" NORMAL, pcsc_stringify_error(rv), rv); \
 else \
 	printf(text ": " BLUE "OK\n\n" NORMAL);
 
@@ -173,10 +173,14 @@ static void parse_properties(unsigned char *bRecvBuffer, int length)
 } /* parse_properties */
 
 
-static const char *pinpad_return_codes(unsigned char bRecvBuffer[])
+static const char *pinpad_return_codes(int length,
+	unsigned char bRecvBuffer[])
 {
 	const char * ret = "UNKNOWN";
 
+	if (length < 2)
+		return "";
+
 	if ((0x90 == bRecvBuffer[0]) && (0x00 == bRecvBuffer[1]))
 		ret = "Success";
 
@@ -240,6 +244,7 @@ int main(int argc, char *argv[])
 #ifdef MODIFY_PIN
 	PIN_MODIFY_STRUCTURE *pin_modify;
 #endif
+	char error;
 	int PIN_min_size = 4;
 	int PIN_max_size = 8;
 
@@ -524,6 +529,8 @@ int main(int argc, char *argv[])
 
 		PCSC_ERROR_CONT(rv, "SCardControl")
 	}
+#else
+	(void)ccid_esc_command;
 #endif
 
 	if (0 == verify_ioctl)
@@ -575,7 +582,7 @@ int main(int argc, char *argv[])
 	}
 
 	/* APDU select applet */
-	printf("Select applet: ");
+	printf("Select applet:");
 	send_length = 11;
 	memcpy(bSendBuffer, "\x00\xA4\x04\x00\x06\xA0\x00\x00\x00\x18\xFF",
 		send_length);
@@ -610,7 +617,7 @@ int main(int argc, char *argv[])
 	pin_verify -> wPINMaxExtraDigit = (PIN_min_size << 8) + PIN_max_size;
 	pin_verify -> bEntryValidationCondition = bEntryValidationCondition;
 	pin_verify -> bNumberMessage = 0x01;
-	pin_verify -> wLangId = 0x0904;
+	pin_verify -> wLangId = 0x0409;		/* United States */
 	pin_verify -> bMsgIndex = 0x00;
 	pin_verify -> bTeoPrologue[0] = 0x00;
 	pin_verify -> bTeoPrologue[1] = 0x00;
@@ -673,14 +680,20 @@ int main(int argc, char *argv[])
 			printf("\n");
 	}
 
-	printf(" card response:");
+	error = FALSE;
+	if (length != 2 || bRecvBuffer[0] != 0x90 || bRecvBuffer[1] != 0x00)
+		error = TRUE;
+
+	printf(error ? RED : GREEN);
+	printf(" card response [%"LF"d bytes]:", length);
 	for (i=0; i<length; i++)
 		printf(" %02X", bRecvBuffer[i]);
-	printf(": %s\n", pinpad_return_codes(bRecvBuffer));
+	printf(": %s", pinpad_return_codes(length, bRecvBuffer));
+	printf(NORMAL "\n");
 	PCSC_ERROR_CONT(rv, "SCardControl")
 
 	/* verify PIN dump */
-	printf("\nverify PIN dump: ");
+	printf("\nverify PIN dump:");
 	send_length = 5;
 	memcpy(bSendBuffer, "\x00\x40\x00\x00\xFF",
 		send_length);
@@ -698,7 +711,7 @@ int main(int argc, char *argv[])
 
 	if ((2 == length) && (0x6C == bRecvBuffer[0]))
 	{
-		printf("\nverify PIN dump: ");
+		printf("\nverify PIN dump:");
 		send_length = 5;
 		memcpy(bSendBuffer, "\x00\x40\x00\x00\xFF",
 			send_length);
@@ -754,7 +767,7 @@ int main(int argc, char *argv[])
 									/* b1 set = current PIN entry requested */
 	pin_modify -> bEntryValidationCondition = bEntryValidationCondition;
 	pin_modify -> bNumberMessage = 0x03; /* see table above */
-	pin_modify -> wLangId = 0x0904;
+	pin_modify -> wLangId = 0x0409;		/* United States */
 	pin_modify -> bMsgIndex1 = 0x00;
 	pin_modify -> bMsgIndex2 = 0x01;
 	pin_modify -> bMsgIndex3 = 0x02;
@@ -830,14 +843,20 @@ int main(int argc, char *argv[])
 		}
 	}
 
-	printf(" card response:");
+	error = FALSE;
+	if (length != 2 || bRecvBuffer[0] != 0x90 || bRecvBuffer[1] != 0x00)
+		error = TRUE;
+
+	printf(error ? RED : GREEN);
+	printf(" card response [%d bytes]:", length);
 	for (i=0; i<length; i++)
 		printf(" %02X", bRecvBuffer[i]);
-	printf("\n");
+	printf(": %s", pinpad_return_codes(length, bRecvBuffer));
+	printf(NORMAL "\n");
 	PCSC_ERROR_CONT(rv, "SCardControl")
 
 	/* modify PIN dump */
-	printf("\nmodify PIN dump: ");
+	printf("\nmodify PIN dump:");
 	send_length = 5;
 	memcpy(bSendBuffer, "\x00\x40\x00\x00\xFF",
 		send_length);
@@ -855,7 +874,7 @@ int main(int argc, char *argv[])
 
 	if ((2 == length) && (0x6C == bRecvBuffer[0]))
 	{
-		printf("\nverify PIN dump: ");
+		printf("\nverify PIN dump:");
 		send_length = 5;
 		memcpy(bSendBuffer, "\x00\x40\x00\x00\xFF",
 			send_length);
